"""
This module contains the structures related to areas of interest.
"""
from dataclasses import dataclass
from typing import NamedTuple, Optional, Tuple, Union

from pyproj.utils import is_null


@dataclass(frozen=True)
class AreaOfInterest:
    """
    .. versionadded:: 2.3.0

    This is the area of interest for:

    - Transformations
    - Querying for CRS data.
    """

    #: The west bound in degrees of the area of interest.
    west_lon_degree: float
    #: The south bound in degrees of the area of interest.
    south_lat_degree: float
    #: The east bound in degrees of the area of interest.
    east_lon_degree: float
    #: The north bound in degrees of the area of interest.
    north_lat_degree: float

    def __post_init__(self):
        if (
            is_null(self.west_lon_degree)
            or is_null(self.south_lat_degree)
            or is_null(self.east_lon_degree)
            or is_null(self.north_lat_degree)
        ):
            raise ValueError("NaN or None values are not allowed.")


class AreaOfUse(NamedTuple):
    """
    .. versionadded:: 2.0.0

    Area of Use for CRS, CoordinateOperation, or a Transformer.
    """

    #: West bound of area of use.
    west: float
    #: South bound of area of use.
    south: float
    #: East bound of area of use.
    east: float
    #: North bound of area of use.
    north: float
    #: Name of area of use.
    name: Optional[str] = None

    @property
    def bounds(self) -> Tuple[float, float, float, float]:
        """
        The bounds of the area of use.

        Returns
        -------
        Tuple[float, float, float, float]
            west, south, east, and north bounds.
        """
        return self.west, self.south, self.east, self.north

    def __str__(self) -> str:
        return f"- name: {self.name}\n" f"- bounds: {self.bounds}"


@dataclass
class BBox:
    """
    Bounding box to check if data intersects/contains other
    bounding boxes.

    .. versionadded:: 3.0.0

    """

    #: West bound of bounding box.
    west: float
    #: South bound of bounding box.
    south: float
    #: East bound of bounding box.
    east: float
    #: North bound of bounding box.
    north: float

    def __post_init__(self):
        if (
            is_null(self.west)
            or is_null(self.south)
            or is_null(self.east)
            or is_null(self.north)
        ):
            raise ValueError("NaN or None values are not allowed.")

    def intersects(self, other: Union["BBox", AreaOfUse]) -> bool:
        """
        Parameters
        ----------
        other: BBox
            The other BBox to use to check.

        Returns
        -------
        bool:
            True if this BBox intersects the other bbox.
        """
        return (
            self.west < other.east
            and other.west < self.east
            and self.south < other.north
            and other.south < self.north
        )

    def contains(self, other: Union["BBox", AreaOfUse]) -> bool:
        """
        Parameters
        ----------
        other: Union["BBox", AreaOfUse]
            The other BBox to use to check.

        Returns
        -------
        bool:
            True if this BBox contains the other bbox.
        """
        return (
            other.west >= self.west
            and other.east <= self.east
            and other.south >= self.south
            and other.north <= self.north
        )



from typing import Optional

def _pyproj_global_context_initialize() -> None: ...
def get_user_data_dir(create: bool = False) -> str: ...
def _global_context_set_data_dir() -> None: ...
def set_use_global_context(active: Optional[bool] = None) -> None: ...



"""
This module is for building coordinate systems to be used when
building a CRS.
"""
from typing import Union

from pyproj._crs import CoordinateSystem
from pyproj.crs.enums import (
    Cartesian2DCSAxis,
    Ellipsoidal2DCSAxis,
    Ellipsoidal3DCSAxis,
    VerticalCSAxis,
)

# useful constants to use when setting PROJ JSON units
UNIT_METRE = "metre"
UNIT_DEGREE = "degree"
UNIT_FT = {"type": "LinearUnit", "name": "foot", "conversion_factor": 0.3048}
UNIT_US_FT = {
    "type": "LinearUnit",
    "name": "US survey foot",
    "conversion_factor": 0.304800609601219,
}

_ELLIPSOIDAL_2D_AXIS_MAP = {
    Ellipsoidal2DCSAxis.LONGITUDE_LATITUDE: [
        {
            "name": "Longitude",
            "abbreviation": "lon",
            "direction": "east",
            "unit": UNIT_DEGREE,
        },
        {
            "name": "Latitude",
            "abbreviation": "lat",
            "direction": "north",
            "unit": UNIT_DEGREE,
        },
    ],
    Ellipsoidal2DCSAxis.LATITUDE_LONGITUDE: [
        {
            "name": "Latitude",
            "abbreviation": "lat",
            "direction": "north",
            "unit": UNIT_DEGREE,
        },
        {
            "name": "Longitude",
            "abbreviation": "lon",
            "direction": "east",
            "unit": UNIT_DEGREE,
        },
    ],
}


class Ellipsoidal2DCS(CoordinateSystem):
    """
    .. versionadded:: 2.5.0

    This generates an Ellipsoidal 2D Coordinate System
    """

    def __new__(
        cls,
        axis: Union[Ellipsoidal2DCSAxis, str] = Ellipsoidal2DCSAxis.LONGITUDE_LATITUDE,
    ):
        """
        Parameters
        ----------
        axis: :class:`pyproj.crs.enums.Ellipsoidal2DCSAxis` or str, optional
            This is the axis order of the coordinate system. Default is
            :attr:`pyproj.crs.enums.Ellipsoidal2DCSAxis.LONGITUDE_LATITUDE`.
        """
        return cls.from_json_dict(
            {
                "type": "CoordinateSystem",
                "subtype": "ellipsoidal",
                "axis": _ELLIPSOIDAL_2D_AXIS_MAP[Ellipsoidal2DCSAxis.create(axis)],
            }
        )


_ELLIPSOIDAL_3D_AXIS_MAP = {
    Ellipsoidal3DCSAxis.LONGITUDE_LATITUDE_HEIGHT: [
        {
            "name": "Longitude",
            "abbreviation": "lon",
            "direction": "east",
            "unit": UNIT_DEGREE,
        },
        {
            "name": "Latitude",
            "abbreviation": "lat",
            "direction": "north",
            "unit": UNIT_DEGREE,
        },
        {
            "name": "Ellipsoidal height",
            "abbreviation": "h",
            "direction": "up",
            "unit": UNIT_METRE,
        },
    ],
    Ellipsoidal3DCSAxis.LATITUDE_LONGITUDE_HEIGHT: [
        {
            "name": "Latitude",
            "abbreviation": "lat",
            "direction": "north",
            "unit": UNIT_DEGREE,
        },
        {
            "name": "Longitude",
            "abbreviation": "lon",
            "direction": "east",
            "unit": UNIT_DEGREE,
        },
        {
            "name": "Ellipsoidal height",
            "abbreviation": "h",
            "direction": "up",
            "unit": UNIT_METRE,
        },
    ],
}


class Ellipsoidal3DCS(CoordinateSystem):
    """
    .. versionadded:: 2.5.0

    This generates an Ellipsoidal 3D Coordinate System
    """

    def __new__(
        cls,
        axis: Union[
            Ellipsoidal3DCSAxis, str
        ] = Ellipsoidal3DCSAxis.LONGITUDE_LATITUDE_HEIGHT,
    ):
        """
        Parameters
        ----------
        axis: :class:`pyproj.crs.enums.Ellipsoidal3DCSAxis` or str, optional
            This is the axis order of the coordinate system. Default is
            :attr:`pyproj.crs.enums.Ellipsoidal3DCSAxis.LONGITUDE_LATITUDE_HEIGHT`.
        """
        return cls.from_json_dict(
            {
                "type": "CoordinateSystem",
                "subtype": "ellipsoidal",
                "axis": _ELLIPSOIDAL_3D_AXIS_MAP[Ellipsoidal3DCSAxis.create(axis)],
            }
        )


_CARTESIAN_2D_AXIS_MAP = {
    Cartesian2DCSAxis.EASTING_NORTHING: [
        {
            "name": "Easting",
            "abbreviation": "E",
            "direction": "east",
            "unit": UNIT_METRE,
        },
        {
            "name": "Northing",
            "abbreviation": "N",
            "direction": "north",
            "unit": UNIT_METRE,
        },
    ],
    Cartesian2DCSAxis.NORTHING_EASTING: [
        {
            "name": "Northing",
            "abbreviation": "N",
            "direction": "north",
            "unit": UNIT_METRE,
        },
        {
            "name": "Easting",
            "abbreviation": "E",
            "direction": "east",
            "unit": UNIT_METRE,
        },
    ],
    Cartesian2DCSAxis.EASTING_NORTHING_FT: [
        {"name": "Easting", "abbreviation": "X", "direction": "east", "unit": UNIT_FT},
        {
            "name": "Northing",
            "abbreviation": "Y",
            "direction": "north",
            "unit": UNIT_FT,
        },
    ],
    Cartesian2DCSAxis.NORTHING_EASTING_FT: [
        {
            "name": "Northing",
            "abbreviation": "Y",
            "direction": "north",
            "unit": UNIT_FT,
        },
        {"name": "Easting", "abbreviation": "X", "direction": "east", "unit": UNIT_FT},
    ],
    Cartesian2DCSAxis.EASTING_NORTHING_US_FT: [
        {
            "name": "Easting",
            "abbreviation": "X",
            "direction": "east",
            "unit": UNIT_US_FT,
        },
        {
            "name": "Northing",
            "abbreviation": "Y",
            "direction": "north",
            "unit": UNIT_US_FT,
        },
    ],
    Cartesian2DCSAxis.NORTHING_EASTING_US_FT: [
        {
            "name": "Northing",
            "abbreviation": "Y",
            "direction": "north",
            "unit": UNIT_US_FT,
        },
        {
            "name": "Easting",
            "abbreviation": "X",
            "direction": "east",
            "unit": UNIT_US_FT,
        },
    ],
    Cartesian2DCSAxis.NORTH_POLE_EASTING_SOUTH_NORTHING_SOUTH: [
        {
            "name": "Easting",
            "abbreviation": "E",
            "direction": "south",
            "unit": UNIT_METRE,
        },
        {
            "name": "Northing",
            "abbreviation": "N",
            "direction": "south",
            "unit": UNIT_METRE,
        },
    ],
    Cartesian2DCSAxis.SOUTH_POLE_EASTING_NORTH_NORTHING_NORTH: [
        {
            "name": "Easting",
            "abbreviation": "E",
            "direction": "north",
            "unit": UNIT_METRE,
        },
        {
            "name": "Northing",
            "abbreviation": "N",
            "direction": "north",
            "unit": UNIT_METRE,
        },
    ],
    Cartesian2DCSAxis.WESTING_SOUTHING: [
        {
            "name": "Easting",
            "abbreviation": "Y",
            "direction": "west",
            "unit": UNIT_METRE,
        },
        {
            "name": "Northing",
            "abbreviation": "X",
            "direction": "south",
            "unit": UNIT_METRE,
        },
    ],
}


class Cartesian2DCS(CoordinateSystem):
    """
    .. versionadded:: 2.5.0

    This generates an Cartesian 2D Coordinate System
    """

    def __new__(
        cls, axis: Union[Cartesian2DCSAxis, str] = Cartesian2DCSAxis.EASTING_NORTHING
    ):
        """
        Parameters
        ----------
        axis: :class:`pyproj.crs.enums.Cartesian2DCSAxis` or str, optional
            This is the axis order of the coordinate system.
            Default is :attr:`pyproj.crs.enums.Cartesian2DCSAxis.EASTING_NORTHING`.
        """
        return cls.from_json_dict(
            {
                "type": "CoordinateSystem",
                "subtype": "Cartesian",
                "axis": _CARTESIAN_2D_AXIS_MAP[Cartesian2DCSAxis.create(axis)],
            }
        )


_VERTICAL_AXIS_MAP = {
    VerticalCSAxis.GRAVITY_HEIGHT: {
        "name": "Gravity-related height",
        "abbreviation": "H",
        "direction": "up",
        "unit": UNIT_METRE,
    },
    VerticalCSAxis.GRAVITY_HEIGHT_US_FT: {
        "name": "Gravity-related height",
        "abbreviation": "H",
        "direction": "up",
        "unit": UNIT_US_FT,
    },
    VerticalCSAxis.GRAVITY_HEIGHT_FT: {
        "name": "Gravity-related height",
        "abbreviation": "H",
        "direction": "up",
        "unit": UNIT_FT,
    },
    VerticalCSAxis.DEPTH: {
        "name": "Depth",
        "abbreviation": "D",
        "direction": "down",
        "unit": UNIT_METRE,
    },
    VerticalCSAxis.DEPTH_US_FT: {
        "name": "Depth",
        "abbreviation": "D",
        "direction": "down",
        "unit": UNIT_US_FT,
    },
    VerticalCSAxis.DEPTH_FT: {
        "name": "Depth",
        "abbreviation": "D",
        "direction": "down",
        "unit": UNIT_FT,
    },
    VerticalCSAxis.UP: {
        "name": "up",
        "abbreviation": "H",
        "direction": "up",
        "unit": UNIT_METRE,
    },
    VerticalCSAxis.UP_FT: {
        "name": "up",
        "abbreviation": "H",
        "direction": "up",
        "unit": UNIT_FT,
    },
    VerticalCSAxis.UP_US_FT: {
        "name": "up",
        "abbreviation": "H",
        "direction": "up",
        "unit": UNIT_US_FT,
    },
}


class VerticalCS(CoordinateSystem):
    """
    .. versionadded:: 2.5.0

    This generates an Vertical Coordinate System
    """

    def __new__(cls, axis: Union[VerticalCSAxis, str] = VerticalCSAxis.GRAVITY_HEIGHT):
        """
        Parameters
        ----------
        axis: :class:`pyproj.crs.enums.VerticalCSAxis` or str, optional
            This is the axis direction of the coordinate system.
            Default is :attr:`pyproj.crs.enums.VerticalCSAxis.GRAVITY_HEIGHT`.
        """
        return cls.from_json_dict(
            {
                "type": "CoordinateSystem",
                "subtype": "vertical",
                "axis": [_VERTICAL_AXIS_MAP[VerticalCSAxis.create(axis)]],
            }
        )



"""
Exceptions for pyproj
"""


class ProjError(RuntimeError):
    """Raised when a Proj error occurs."""

    internal_proj_error = None

    def __init__(self, error_message: str) -> None:
        if self.internal_proj_error is not None:
            error_message = (
                f"{error_message}: (Internal Proj Error: {self.internal_proj_error})"
            )
            ProjError.clear()
        super().__init__(error_message)

    @staticmethod
    def clear() -> None:
        """
        This will clear the internal PROJ error message.
        """
        ProjError.internal_proj_error = None


class CRSError(ProjError):
    """Raised when a CRS error occurs."""


class GeodError(RuntimeError):
    """Raised when a Geod error occurs."""


class DataDirError(RuntimeError):
    """Raised when a the data directory was not found."""
